# coding=utf-8
# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import (absolute_import, division, generators, nested_scopes, print_function,
                        unicode_literals, with_statement)

from pants.backend.jvm.tasks.nailgun_task import NailgunTask
from pants.base.exceptions import TaskError
from pants.base.workunit import WorkUnitLabel
from pants.option.ranked_value import RankedValue

from pants.contrib.scrooge.tasks.thrift_util import calculate_compile_sources


class ThriftLintError(Exception):
  """Raised on a lint failure."""


class ThriftLinter(NailgunTask):
  """Print linter warnings for thrift files."""

  @staticmethod
  def _is_thrift(target):
    return target.is_thrift

  @classmethod
  def register_options(cls, register):
    super(ThriftLinter, cls).register_options(register)
    register('--skip', type=bool, fingerprint=True, help='Skip thrift linting.')
    register('--strict', type=bool, fingerprint=True,
             help='Fail the goal if thrift linter errors are found. Overrides the '
                  '`strict-default` option.')
    register('--strict-default', default=False, advanced=True, type=bool,
             fingerprint=True,
             help='Sets the default strictness for targets. The `strict` option overrides '
                  'this value if it is set.')
    register('--linter-args', default=[], advanced=True, type=list, fingerprint=True,
             help='Additional options passed to the linter.')
    cls.register_jvm_tool(register, 'scrooge-linter')

  @classmethod
  def product_types(cls):
    # Declare the product of this goal. Gen depends on thrift-linter.
    return ['thrift-linter']

  @property
  def cache_target_dirs(self):
    return True

  @staticmethod
  def _to_bool(value):
    # Converts boolean and string values to boolean.
    return str(value) == 'True'

  def _is_strict(self, target):
    # The strict value is read from the following, in order:
    # 1. the option --[no-]strict, but only if explicitly set.
    # 2. java_thrift_library target in BUILD file, thrift_linter_strict = False,
    # 3. options, --[no-]strict-default
    options = self.get_options()
    if options.get_rank('strict') > RankedValue.HARDCODED:
      return self._to_bool(self.get_options().strict)

    if target.thrift_linter_strict is not None:
      return self._to_bool(target.thrift_linter_strict)

    return self._to_bool(self.get_options().strict_default)

  def _lint(self, target):
    self.context.log.debug('Linting {0}'.format(target.address.spec))

    classpath = self.tool_classpath('scrooge-linter')
    config_args = []

    config_args.extend(self.get_options().linter_args)
    if not self._is_strict(target):
      config_args.append('--ignore-errors')

    include_paths , paths = calculate_compile_sources([target], self._is_thrift)
    for p in include_paths:
      config_args.extend(['--include-path', p])

    args = config_args + list(paths)


    # If runjava returns non-zero, this marks the workunit as a
    # FAILURE, and there is no way to wrap this here.
    returncode = self.runjava(classpath=classpath,
                              main='com.twitter.scrooge.linter.Main',
                              args=args,
                              jvm_options=self.get_options().jvm_options,
                              workunit_labels=[WorkUnitLabel.COMPILER])  # to let stdout/err through.

    if returncode != 0:
      raise ThriftLintError(
        'Lint errors in target {0} for {1}.'.format(target.address.spec, paths))

  def execute(self):
    if self.get_options().skip:
      return

    thrift_targets = self.context.targets(self._is_thrift)
    with self.invalidated(thrift_targets) as invalidation_check:
      errors = []
      for vt in invalidation_check.invalid_vts:
        try:
          self._lint(vt.target)
        except ThriftLintError as e:
          errors.append(str(e))
        else:
          vt.update()
      if errors:
        raise TaskError('\n'.join(errors))



# coding=utf-8
# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import (absolute_import, division, generators, nested_scopes, print_function,
                        unicode_literals, with_statement)

import os
import re


INCLUDE_PARSER = re.compile(r'^\s*include\s+"([^"]+)"\s*([\/\/|\#].*)*$')


def find_includes(basedirs, source, log=None):
  """Finds all thrift files included by the given thrift source.

  :basedirs: A set of thrift source file base directories to look for includes in.
  :source: The thrift source file to scan for includes.
  :log: An optional logger
  """

  all_basedirs = [os.path.dirname(source)]
  all_basedirs.extend(basedirs)

  includes = set()
  with open(source, 'r') as thrift:
    for line in thrift.readlines():
      match = INCLUDE_PARSER.match(line)
      if match:
        capture = match.group(1)
        added = False
        for basedir in all_basedirs:
          include = os.path.join(basedir, capture)
          if os.path.exists(include):
            if log:
              log.debug('{} has include {}'.format(source, include))
            includes.add(include)
            added = True
        if not added:
          raise ValueError("{} included in {} not found in bases {}"
                           .format(include, source, all_basedirs))
  return includes


def find_root_thrifts(basedirs, sources, log=None):
  """Finds the root thrift files in the graph formed by sources and their recursive includes.

  :basedirs: A set of thrift source file base directories to look for includes in.
  :sources: Seed thrift files to examine.
  :log: An optional logger.
  """

  root_sources = set(sources)
  for source in sources:
    root_sources.difference_update(find_includes(basedirs, source, log=log))
  return root_sources


def calculate_compile_sources(targets, is_thrift_target):
  """Calculates the set of thrift source files that need to be compiled.
  It does not exclude sources that are included in other sources.

  A tuple of (include basedirs, thrift sources) is returned.

  :targets: The targets to examine.
  :is_thrift_target: A predicate to pick out thrift targets for consideration in the analysis.
  """

  basedirs = set()
  sources = set()
  def collect_sources(target):
    basedirs.add(target.target_base)
    sources.update(target.sources_relative_to_buildroot())

  for target in targets:
    target.walk(collect_sources, predicate=is_thrift_target)
  return basedirs, sources




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: ContribPlugin(BuildFileAddress(BuildFile(contrib/scrooge/src/python/pants/contrib/scrooge/BUILD, FileSystemProjectTree(/Users/stuhood/src/pants)), plugin))

from setuptools import setup

setup(**
{   'classifiers': [   'Intended Audience :: Developers',
                       'License :: OSI Approved :: Apache Software License',
                       'Operating System :: MacOS :: MacOS X',
                       'Operating System :: POSIX :: Linux',
                       'Programming Language :: Python',
                       'Topic :: Software Development :: Build Tools',
                       'Topic :: Software Development :: Code Generators'],
    'description': 'Scrooge thrift generator pants plugins.',
    'entry_points': {   'pantsbuild.plugin': [   'register_goals = pants.contrib.scrooge.register:register_goals']},
    'install_requires': [   'twitter.common.collections>=0.3.1,<0.4',
                            'pantsbuild.pants==1.2.0rc2'],
    'license': 'Apache License, Version 2.0',
    'long_description': 'Pants is an Apache2 licensed build tool written in Python.\n\nThe latest documentation can be found `here <http://pantsbuild.github.io/>`_.\n\n1.2.x Stable Releases\n=====================\n\nThis document describes releases from the ``1.2.x`` ``stable`` branch.\n\n\n1.2.0rc2 (10/18/2016)\n---------------------\n\nThe third, and perhaps final, release candidate for stable 1.2.0.\n\nNotably: this release candidate restores JDK7 compatibility for junit tests.\n\nAPI Changes\n~~~~~~~~~~~\n\n* Update junit runner to 1.0.15 to get java 7 compatibility\n  `RB #4324 <https://rbcommons.com/s/twitter/r/4324>`_\n\n* Back down the minimum required java version for running Pants tools to java 7\n  `RB #4314 <https://rbcommons.com/s/twitter/r/4314>`_\n\nBugfixes\n~~~~~~~~\n\n* Fix exclude_target_regexp breakage in test-changed and --files option breakage in changed with diffspec\n  `RB #4321 <https://rbcommons.com/s/twitter/r/4321>`_\n\n* Prevent cleanup error at end of pants test with --test-junit-html-report option, update safe_rmtree to be symlink aware\n  `RB #4319 <https://rbcommons.com/s/twitter/r/4319>`_\n\n* fix --changed-files option\n  `RB #4309 <https://rbcommons.com/s/twitter/r/4309>`_\n\nNew Features\n~~~~~~~~~~~~\n\n* Adding scalafmt formatting to fmt goal and enable by default\n  `RB #4312 <https://rbcommons.com/s/twitter/r/4312>`_\n\nRefactoring, Improvements, and Tooling\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n* migrate changed integration tests to isolated temp git repos and add an environment variable to override buildroot\n  `RB #4295 <https://rbcommons.com/s/twitter/r/4295>`_\n\n1.2.0rc1 (10/13/2016)\n---------------------\n\nSecond release candidate for stable 1.2.0.\n\nAPI Changes\n~~~~~~~~~~~\n\n* Add subsystem_utils to test_infra\n  `RB #4303 <https://rbcommons.com/s/twitter/r/4303>`_\n\nBugfixes\n~~~~~~~~\n\n* Switch default deference back to True for tarball artifact\n  `RB #4304 <https://rbcommons.com/s/twitter/r/4304>`_\n\n* Filter inactive goals from `Goal.all`.\n  `RB #4298 <https://rbcommons.com/s/twitter/r/4298>`_\n\n* JUnit runner fix for len(args) > max_args in argfile.safe_args\n  `RB #4294 <https://rbcommons.com/s/twitter/r/4294>`_\n\nRefactoring, Improvements, and Tooling\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n* Version clarification\n  `RB #4299 <https://rbcommons.com/s/twitter/r/4299>`_\n\n1.2.0rc0 (10/07/2016)\n---------------------\n\nFirst release candidate for stable 1.2.0.\n\nNew Features\n~~~~~~~~~~~~\n\n* Make the name= target keyword optional in BUILD files.\n  `RB #4275 <https://rbcommons.com/s/twitter/r/4275>`_\n\n* Add Scalafmt Support to Pants\n  `RB #4216 <https://rbcommons.com/s/twitter/r/4216>`_\n\nRefactoring, Improvements, and Tooling\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n* Add libffi to pants pre-reqs, qualify JDK req.\n  `RB #4285 <https://rbcommons.com/s/twitter/r/4285>`_\n\n* Update the setup.py description.\n  `RB #4283 <https://rbcommons.com/s/twitter/r/4283>`_\n\n* Refactor Intermediate Target Generation Logic\n  `RB #4277 <https://rbcommons.com/s/twitter/r/4277>`_\n\n* Clean up after failed artifact extractions\n  `RB #4255 <https://rbcommons.com/s/twitter/r/4255>`_\n\n* Publish the CPP plugin\n  `RB #4282 <https://rbcommons.com/s/twitter/r/4282>`_\n\n* Change --no-dryrun to the new flag in docs\n  `RB #4280 <https://rbcommons.com/s/twitter/r/4280>`_\n\n* Add --no-transitive flag to findbugs so you can run findbugs only for the targets specified on the command line\n  `RB #4276 <https://rbcommons.com/s/twitter/r/4276>`_\n\nNew Engine Work\n~~~~~~~~~~~~~~~\n\n* [engine] Rule Graph construction perf improvements\n  `RB #4281 <https://rbcommons.com/s/twitter/r/4281>`_\n\n* [engine] Introduce static analysis model and replace validator with it\n  `RB #4251 <https://rbcommons.com/s/twitter/r/4251>`_\n',
    'name': 'pantsbuild.pants.contrib.scrooge',
    'namespace_packages': ['pants', 'pants.contrib'],
    'package_data': {   },
    'package_dir': {   '': 'src'},
    'packages': [   'pants',
                    'pants.contrib',
                    'pants.contrib.scrooge',
                    'pants.contrib.scrooge.tasks'],
    'url': 'https://github.com/pantsbuild/pants',
    'version': '1.2.0rc2',
    'zip_safe': True}
)



# coding=utf-8
# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import (absolute_import, division, generators, nested_scopes, print_function,
                        unicode_literals, with_statement)

import hashlib

from pants.backend.codegen.targets.java_thrift_library import JavaThriftLibrary
from pants.base.fingerprint_strategy import FingerprintStrategy


class JavaThriftLibraryFingerprintStrategy(FingerprintStrategy):
  """Scrooge cares about a thrift target's `language` and `rpc_style` in addition to its payload.

  As such this strategy ensures new code is generated by scrooge whenever `language` or `rpc_style`
  changes.
  """

  def __init__(self, thrift_defaults):
    self._thrift_defaults = thrift_defaults

  def compute_fingerprint(self, target):
    fp = target.payload.fingerprint()
    if not isinstance(target, JavaThriftLibrary):
      return fp

    hasher = hashlib.sha1()
    hasher.update(fp)
    hasher.update(self._thrift_defaults.language(target))
    hasher.update(self._thrift_defaults.rpc_style(target))
    return hasher.hexdigest()

  def __hash__(self):
    return hash((type(self), self._thrift_defaults))

  def __eq__(self, other):
    return type(self) == type(other) and self._thrift_defaults == other._thrift_defaults
